-- Another wonderful Wrk production

---- Configurable stuff

-- TimeLimit: Time, in game ticks before the game will end. Set to nil to disable the time limit
-- Example: 30 * 60 * 3 for 3 minutes
TimeLimit = nil

-- SlidingBots: If true, bots will move around while shooting to make them act more like real players.
-- It looks a bit goofy, but makes it more challenging
SlidingBots = true

-- NumBots: The game will attempt to populate the map with this number of bots
NumBots = 5

-- ForceWeapons: The player and bots will only be able to use the specified weapon
-- Options are: nil, "fist", "pistol", "fusion pistol", "assault rifle", "missile launcher", "alien weapon", "flamethrower", "shotgun", or "smg".
-- Set to nil if you don't want to enable this option
ForceWeapons = nil

-- RestrictBotWeapons: If true, only bots with the weapons available on the map will spawn.
-- Otherwise, bots with any type of weapon can spawn
RestrictBotWeapons = true

-- EMFH: If true, all bots will fight each other.  If false, it's everyone vs you.
EMFH = true

-- AllowMADDs: If true, MADDs can spawn.  Otherwise, they cannot.
AllowMADDs = false

-- Game.monsters_replenish: set to false if you don't want regular non-bot monsters spawning, true if you do.
Game.monsters_replenish = false

-- BRUTAL: If true, bots explode into gore on death
BRUTAL = false

-- You can assign weights to certain weapon types.  A higher number means bots of that type will appear more often.
-- These defaults are chosen based on what weapons players are actually likely to use in a real netgame.
WEIGHT_FIST = 1
WEIGHT_MADD = 4
WEIGHT_PISTOL = 4
WEIGHT_ROCKET = 30
WEIGHT_FUSION = 8
WEIGHT_SHOTGUN = 10
WEIGHT_ASSAULT_RIFLE = 10
WEIGHT_SMG = 15
WEIGHT_ALIEN_WEAPON = 8
WEIGHT_FLAMETHROWER = 4




-- Don't modify below this

Triggers = {}
Starts = {}
NumBotSpawnPolys = 0
BotSpawnPolys = {}
NumAvailableBotTypes = 0
AvailableBotTypes = {}
Game.proper_item_accounting = true
CollectionsUsed = {3}
moving_ephemera = {}
SP_NETGAME_OWNER = "SP NETGAME"
MAX_RENDERED_COUNT = 1400

function Triggers.init()
    for start in PlayerStarts() do
        table.insert(Starts, start)

        if not is_bot_spawn_poly(start.polygon) then
            add_bot_spawn_poly(start.polygon)
        end
    end
    for p in Players() do
        p._deaths = 0
        p._kills = 0
    end

    if not EMFH then
        -- Everyone vs the player
        for m in MonsterTypes() do
            for c in MonsterClasses() do
                if c == "player" then
                    m.friends[c] = false
                    m.enemies[c] = true
                else
                    if c == FriendClass then
                        m.friends[c] = false
                        m.enemies[c] = true
                    else
                        m.friends[c] = true
                        m.enemies[c] = false
                    end
                end
            end
        end
    end

    if ForceWeapons then
        add_available_bot_type(get_bot_type_from_item(ForceWeapons), 1)
    elseif RestrictBotWeapons then
        -- Allows allow punching dudes
        add_available_bot_type("minor tick", WEIGHT_FIST)

        if AllowMADDs then
            add_available_bot_type("tiny yeti", WEIGHT_MADD)
        end

        for item in ItemStarts() do
            local bot_type = get_bot_type_from_item(item.type)
            if bot_type and not is_existing_available_bot_type(bot_type) then
                add_available_bot_type(bot_type, get_weighted_value(item.type))
            end
        end
    else
        add_available_bot_type("major tick", get_weighted_value("pistol"))
        add_available_bot_type("minor cyborg", get_weighted_value("missile launcher"))
        add_available_bot_type("minor tick", get_weighted_value("knife"))
        add_available_bot_type("big minor drone", get_weighted_value("fusion pistol"))
        add_available_bot_type("big major drone", get_weighted_value("shotgun"))
        add_available_bot_type("kamikaze tick", get_weighted_value("assault rifle"))
        add_available_bot_type("major cyborg", get_weighted_value("smg"))
        add_available_bot_type("tiny pfhor", get_weighted_value("alien weapon"))
        add_available_bot_type("possessed drone", get_weighted_value("flamethrower"))

        if AllowMADDs then
            add_available_bot_type("tiny yeti", WEIGHT_MADD)
        end
    end

    -- Place bots on the map initially based on player start location polygons
    local num = math.min(NumBots, NumBotSpawnPolys)
    for i = 1, num do
        spawn_bot()
    end

    -- Allow bots to use platforms
    for plat in Platforms() do
        if plat.player_controllable then
            plat.monster_controllable = true
        end
    end

    -- Gore stuff
    ephemera_count = 0
    max_ephemera = math.ceil(#Ephemera / 2) -- Leave room for other types of ephemera
end

function spawn_bot()
    local poly = select_random_bot_poly()
    local type = get_random_bot_type()
    local m = Monsters.new(poly.x, poly.y, 0, poly, type)
    if m then
        m.active = true
    end
end

function Triggers.idle()
    for p in Players() do
        if not time_expired() then
            if p.totally_dead and p.action_flags.action_trigger then
                p:revive()
                p.action_flags.action_trigger = false
                p._dead = false
            elseif not p.dead and ForceWeapons then
                p.items["pistol"] = 2
                p.items["pistol ammo"] = 5
                p.items["fusion pistol"] = 1
                p.items["fusion pistol ammo"] = 5
                p.items["assault rifle"] = 1
                p.items["assault rifle ammo"] = 5
                p.items["assault rifle grenades"] = 5
                p.items["missile launcher"] = 1
                p.items["missile launcher ammo"] = 5
                p.items["alien weapon"] = 1
                p.items["flamethrower"] = 1
                p.items["flamethrower ammo"] = 5
                p.items["shotgun"] = 2
                p.items["shotgun ammo"] = 5
                p.items["smg"] = 1
                p.items["smg ammo"] = 5
                p.weapons[ForceWeapons]:select()
            end

            -- Add polygons the player has touched to the possible places bots can spawn
            if p.monster.polygon.type == "normal" and not p.monster.polygon.media then
                if not is_bot_spawn_poly(p.monster.polygon) then
                    add_bot_spawn_poly(p.monster.polygon)
                end
            end
        else -- Time expired
            p.action_flags.action_trigger = false
            p.action_flags.cycle_weapons_backward = false
            p.action_flags.cycle_weapons_forward = false
            p.action_flags.left_trigger = false
            p.action_flags.right_trigger = false
            p.action_flags.microphone_button = false
            p.action_flags.toggle_map = false

            if not p._end_life then
                p._end_life = p.life
            end
            p.life = p._end_life

            p.weapons.active = false

            Level.fog.active = true
            Level.fog.color.r = 0
            Level.fog.color.g = 0
            Level.fog.color.b = 0
            Level.fog.depth = 0
        end

        -- Overlay stuff
        local score = p._kills - p._deaths
        p.overlays[0].color = "yellow"
        p.overlays[1].color = "green"
        p.overlays[2].color = "red"
        p.overlays[3].color = "white"

        if score > 0 then
            p.overlays[0].text = "Score: +" .. score
        else
            p.overlays[0].text = "Score: " .. score
        end
        p.overlays[1].text = "Kills: " .. p._kills
        p.overlays[2].text = "Deaths: " .. p._deaths
        p.overlays[3].text = sec_to_time(Game.ticks / 30)
    end

    local num_bots = 0
    for m in Monsters() do
        if not m.player then
            if not time_expired() then
                -- Try to keep them doing stuff
                m.active = true
                if Game.ticks % 90 == 0 then
                    if m._x then
                        if m.x == m._x and m.y == m._y and m.yaw == m._yaw then
                            m:damage(m.life + 1) -- Kill stalled bots
                        end
                    end
                    m._x = m.x
                    m._y = m.y
                    m._yaw = m.yaw
                end

                if is_bot_monster(m.type) then
                    num_bots = num_bots + 1
                end

                -- Add polygons the monster has touched to the possible places bots can spawn
                if m.valid and m.life > 0 then
                    if not is_bot_spawn_poly(m.polygon) then
                        add_bot_spawn_poly(m.polygon)
                    end
                end

                -- Make the monsters move around while shooting, to make them act more like real players
                if SlidingBots then
                    if is_bot_monster(m.type) and m.action == "attacking far" or m.action == "attacking close" then
                        if m.type.mnemonic == "possessed drone" then
                            x, y, poly = get_offset_location(m.x, m.y, m.z, m.polygon, m.facing, .11)
                            if poly then
                                m:position(x, y, m.z, poly)
                            end
                        else
                            local polys = {}
                            if m._target_poly then
                                table.insert(polys, m._target_poly)
                            else
                                for adj in m.polygon:adjacent_polygons() do
                                    if
                                        math.abs(adj.floor.height - m.polygon.floor.height) <= .3 and
                                            adj.ceiling.height - adj.floor.height >= 1
                                     then
                                        table.insert(polys, adj)
                                    end
                                end
                            end

                            if #polys > 0 then
                                local rand = Game.random(#polys) + 1
                                local target_poly = polys[rand]
                                m._target_poly = target_poly
                                local angle = angle_between_points(m, target_poly)
                                local distance = get_distance(m, target_poly)

                                if distance < .1 then
                                    m._target_poly = nil
                                end

                                local dist_x = distance * math.cos(angle)
                                local dist_y = distance * math.sin(angle)
                                local x_ratio = math.abs(dist_x) / (math.abs(dist_x) + math.abs(dist_y))
                                local y_ratio = math.abs(dist_y) / (math.abs(dist_x) + math.abs(dist_y))

                                local max_dist = .11
                                local move_x = max_dist * x_ratio
                                local move_y = max_dist * y_ratio
                                if target_poly.x < m.x then
                                    move_x = move_x * -1
                                end
                                if target_poly.y < m.y then
                                    move_y = move_y * -1
                                end

                                local new_x = m.x + move_x
                                local new_y = m.y + move_y
                                if m.polygon:contains(new_x, new_y) then
                                    m:position(new_x, new_y, m.z, m.polygon)
                                elseif target_poly:contains(new_x, new_y) then
                                    m:position(new_x, new_y, m.z, target_poly)
                                else
                                    m._target_poly = nil
                                end
                            end
                        end
                    else
                        m._target_poly = nil
                    end
                end
            else -- Time has expired
                m:damage(m.life + 1)
            end
        end
    end

    if not time_expired() and num_bots < NumBots then
        spawn_bot()
    end

    -- Gore stuff
    if BRUTAL then
        for i = #moving_ephemera, 1, -1 do
            local e = update_ephemera(moving_ephemera[i])
            if not e then
                table.remove(moving_ephemera, i)
            end
        end

        -- Try to keep frame rate high
        if Game.ticks % 30 == 0 then
            cleanup_ephemera_in_view()
        end
    end
end

function time_expired()
    if not TimeLimit then
        return false
    end

    return Game.ticks >= TimeLimit
end

function Triggers.player_killed(player, aggressor_player, action, projectile)
    -- Gore stuff
    if BRUTAL then
        create_explosion(player.monster, projectile)
    end
end

function Triggers.monster_killed(monster, aggressor_player, projectile)
    if projectile and projectile.owner and projectile.owner.player then
        projectile.owner.player._kills = projectile.owner.player._kills + 1
        projectile.owner.player:print("Killed " .. get_monster_name(monster.type))
    end

    -- Gore stuff
    if BRUTAL then
        create_explosion(monster, projectile)
    end
end

function create_explosion(monster, projectile)
    if projectile then
        local explode = 0
        if
            projectile.type == "grenade" or projectile.type == "compiler bolt minor" or
                projectile.type == "compiler bolt major" or
                projectile.type == "fusion bolt major" or
                projectile.type == "juggernaut rocket" or
                projectile.type == "trooper grenade" or
                projectile.type == "juggernaut missile" or
                projectile.type == "overloaded fusion disperal"
         then
            explode = 1
        end
        if projectile.type == "missile" then
            explode = 2
        end
        if projectile.type == "flamethrower burst" or projectile.type == "alien weapon" then
            explode = 3
        end
        local direction = angle_between_points(projectile, monster)
        explode_monster(monster, explode, direction)
    else
        explode = 1
        explode_monster(monster, explode)
    end
end

function Triggers.player_damaged(victim, aggressor_player, aggressor_monster, damage_type, damage_amount, projectile)
    if victim.life < 0 then
        if not victim._dead then -- Just died
            victim._deaths = victim._deaths + 1
            victim._dead = true
            if aggressor_monster then
                victim:print("Killed by " .. get_monster_name(aggressor_monster.type))
            else
                victim:print("Killed")
            end
        end
    end

    if BRUTAL then
        spill_blood(victim.monster, projectile)
    end
end

function Triggers.monster_damaged(monster, aggressor_monster, damage_type, damage_amount, projectile)
    if
        not monster.player and monster.life < 0 and projectile and projectile.type == "missile" and
            is_bot_monster(monster.type)
     then
        monster.external_velocity = .15
        monster.vertical_velocity = .15
        monster.facing = angle_between_points(monster, projectile)
    end

    -- Gore stuff
    if BRUTAL then
        if is_bot_monster(monster.type) and monster.type.mnemonic ~= "tiny yeti" then
            spill_blood(monster, projectile)
        end
    end
end

function Triggers.projectile_created(projectile)
    if projectile.owner and not projectile.owner.player and is_bot_monster(projectile.owner.type) then
        if projectile.type == "missile" and projectile.target and projectile.z > projectile.target.z then
            local dist = get_distance(projectile, projectile.target)
            local height = math.abs(projectile.z - projectile.target.z + .02)
            local angle = math.atan2(height, dist)

            angle = -angle * (180 / 3.1415)
            if angle < -30 then
                angle = -30
            end
            projectile.pitch = angle
        end
        local angle = angle_between_points(projectile, projectile.target)
        angle = angle + (Game.random(3)) - 1
        projectile.yaw = angle
    end
end

function sec_to_time(sec)
    local sec = tonumber(sec)

    if sec <= 0 then
        return "00:00:00"
    else
        hours = string.format("%02.f", math.floor(sec / 3600))
        mins = string.format("%02.f", math.floor(sec / 60 - (hours * 60)))
        secs = string.format("%02.f", math.floor(sec - hours * 3600 - mins * 60))
        return hours .. ":" .. mins .. ":" .. secs
    end
end

function get_monster_name(type)
    if type.mnemonic == "major tick" then
        return "red dude"
    elseif type.mnemonic == "minor cyborg" then
        return "orange dude"
    elseif type.mnemonic == "minor tick" then
        return "orange dude"
    elseif type.mnemonic == "tiny yeti" then
        return "a MADD"
    elseif type.mnemonic == "big minor drone" then
        return "yellow dude"
    elseif type.mnemonic == "big major drone" then
        return "green dude"
    elseif type.mnemonic == "kamikaze tick" then
        return "white dude"
    elseif type.mnemonic == "major cyborg" then
        return "slate dude"
    elseif type.mnemonic == "tiny pfhor" then
        return "violet dude"
    elseif type.mnemonic == "possessed drone" then
        return "blue dude"
    else
        return type.mnemonic
    end
end

function is_bot_spawn_poly(poly)
    for _, value in pairs(BotSpawnPolys) do
        if value == poly then
            return true
        end
    end
    return false
end

function add_bot_spawn_poly(poly)
    BotSpawnPolys[NumBotSpawnPolys] = poly
    NumBotSpawnPolys = NumBotSpawnPolys + 1
end

function is_existing_available_bot_type(type)
    for _, value in pairs(AvailableBotTypes) do
        if value == type then
            return true
        end
    end
    return false
end

function add_available_bot_type(type, weight)
    for i = 1, weight do
        AvailableBotTypes[NumAvailableBotTypes] = type
        NumAvailableBotTypes = NumAvailableBotTypes + 1
    end
end

function get_bot_type_from_item(type)
    if type == "pistol" then
        return "major tick"
    elseif type == "fusion pistol" then
        return "big minor drone"
    elseif type == "assault rifle" then
        return "kamikaze tick"
    elseif type == "missile launcher" then
        return "minor cyborg"
    elseif type == "alien weapon" then
        return "tiny pfhor"
    elseif type == "flamethrower" then
        return "possessed drone"
    elseif type == "shotgun" then
        return "big major drone"
    elseif type == "smg" then
        return "major cyborg"
    elseif type == "fist" then
        return "minor tick"
    end

    return nil
end

function select_random_bot_poly()
    local rand = Game.random(NumBotSpawnPolys)
    local poly = BotSpawnPolys[rand]

    -- Try not to select a location really close to the player
    local attempts = 0
    while get_distance(Players[0], poly) < 5 and attempts < 5 do
        rand = Game.random(NumBotSpawnPolys)
        poly = BotSpawnPolys[rand]
        attempts = attempts + 1
    end

    return poly
end

function is_bot_monster(type)
    if
        type == "major tick" or type == "minor cyborg" or type == "minor tick" or type == "big minor drone" or
            type == "big major drone" or
            type == "kamikaze tick" or
            type == "major cyborg" or
            type == "tiny pfhor" or
            type == "possessed drone" or
            type == "tiny yeti"
     then
        return true
    end
    return false
end

function get_random_bot_type()
    local rand = Game.random(NumAvailableBotTypes)
    local type = AvailableBotTypes[rand]
    return type
end

function get_weighted_value(type)
    if type == "pistol" then
        return WEIGHT_PISTOL
    elseif type == "missile launcher" then
        return WEIGHT_ROCKET
    elseif type == "fusion pistol" then
        return WEIGHT_FUSION
    elseif type == "shotgun" then
        return WEIGHT_SHOTGUN
    elseif type == "assault rifle" then
        return WEIGHT_ASSAULT_RIFLE
    elseif type == "smg" then
        return WEIGHT_SMG
    elseif type == "alien weapon" then
        return WEIGHT_ALIEN_WEAPON
    elseif type == "flamethrower" then
        return WEIGHT_FLAMETHROWER
    else
        return 1
    end
end

function get_distance(p1, p2)
    return math.sqrt((p1.x - p2.x) ^ 2 + (p1.y - p2.y) ^ 2)
end

function get_offset_location(x, y, z, polygon, angle, dist)
    local nx, ny = calc_offset_location(x, y, angle, dist)
    local poly = polygon:find_polygon(x, y, nx, ny)

    -- Don't move into polygons if the floor or ceiling block the movement
    if poly and (z <= poly.floor.height or z >= poly.ceiling.height) then
        poly = nil
    end

    return nx, ny, poly
end

function calc_offset_location(x, y, angle, dist)
    angle = angle * math.pi / 180
    return x + math.cos(angle) * dist, y + math.sin(angle) * dist
end

---- The following functions taken from
---- https://gist.github.com/kirubz/fa84375008d376a2d695618e0ae3aed8/9812f1407136ca1ef795d484162d05ad2c7eb2b8
function angleOfPoint(pt)
    local x, y = pt.x, pt.y
    local radian = math.atan2(y, x)
    local angle = radian * 180 / math.pi
    if angle < 0 then
        angle = 360 + angle
    end
    return angle
end

-- returns the degrees between two points (note: 0 degrees is 'east')
function angle_between_points(a, b)
    local x, y = b.x - a.x, b.y - a.y
    return angleOfPoint({x = x, y = y})
end
----

-- Functions for gore stuff

function spill_blood(monster, projectile)
    local m = monster
    local col, seq, ct = 6, 38, 0

    if not col or (not m.player and m.life < 0) then
        return
    end

    local num = 1
    if Ephemera.quality == "ultra" then
        num = 3
    elseif Ephemera.quality == "high" then
        num = 2
    end

    if
        projectile and
            (projectile.type == "missile" or projectile.type == "grenade" or projectile.type == "compiler bolt major" or
                projectile.type == "fusion bolt major" or
                projectile.type == "juggernaut rocket" or
                projectile.type == "trooper grenade" or
                projectile.type == "juggernaut missile" or
                projectile.type == "minor cyborg ball" or
                projectile.type == "major cyborg ball")
     then
        num = num * 3
    end

    for i = 1, num do
        local facing = Game.random_local(360)
        local e = create_ephemera(m.x, m.y, m.z + .4 + Game.random_local(3) / 10.0, m.polygon, col, seq, facing)

        if e then
            table.insert(moving_ephemera, e)

            e._horizontal_dist = Game.random_local(30) / 750.0
            e._vertical_velocity = .02
            e._vertical_velocity_step = -.002
            e._terminal_velocity = -.09
            e._max_bounces = 0
            e._bounces = 0
            e._moving_horizontally = true
            e._active = true
            e.clut_index = ct
        end
    end
end

function explode_monster(m, explode, proj_facing)
    if Ephemera.quality == "off" then
        return
    end

    local col, seq, ct, num = get_ephemera_by_monster(m)

    if Ephemera.quality == "ultra" then
        num = num * 4
    elseif Ephemera.quality == "high" then
        num = num * 2
    elseif Ephemera.quality == "low" then
        num = math.ceil(num * .5)
    end

    if col and seq and ct then
        for i = 1, num do
            if explode == 3 then
                col, seq, ct = get_ephemera_by_flaming_death(m)
            else
                col, seq, ct, num = get_ephemera_by_monster(m)
            end
            local facing = nil
            if proj_facing then
                local rand = -45 + Game.random_local(90)
                facing = proj_facing + rand
            else
                facing = Game.random_local(360)
            end
            local e = create_ephemera(m.x, m.y, m.z + .2 + Game.random_local(5) / 10.0, m.polygon, col, seq, facing)

            if e then
                table.insert(moving_ephemera, e)

                e._horizontal_dist = Game.random_local(100) / 750.0
                if explode == 1 then
                    e._vertical_velocity = .05 + (Game.random_local(5) / 100.0)
                elseif explode == 2 then
                    e._vertical_velocity = .08 + (Game.random_local(5) / 100.0)
                    e._horizontal_dist = (50 + Game.random_local(50)) / 300.0
                elseif explode == 3 then
                    e._vertical_velocity = .02
                else
                    e._vertical_velocity = .01
                end
                e._vertical_velocity_step = -.002
                e._terminal_velocity = -.09
                e._max_bounces = 2
                e._bounces = 0
                e._moving_horizontally = true
                e._active = true
                e.clut_index = ct
            end
        end
    end
end

function update_ephemera(e)
    if not e._active then
        return nil
    end

    if e.z > e.polygon.floor.z or e._vertical_velocity > 0 then
        local vv = e._vertical_velocity + e._vertical_velocity_step
        if vv < e._terminal_velocity then
            vv = e._terminal_velocity
        end
        e._vertical_velocity = vv
        local z = e.z + e._vertical_velocity

        if z > e.polygon.ceiling.z then
            z = e.polygon.ceiling.z
        elseif z < e.polygon.floor.z then
            z = e.polygon.floor.z
        end

        local x = e.x
        local y = e.y
        local poly = e.polygon
        x, y, poly = get_offset_location(e.x, e.y, e.z, e.polygon, e.facing, e._horizontal_dist)

        -- If hit a wall, try picking a new random direction to move in
        if not poly then
            local chdir = 90 + Game.random_local(45)
            if Game.random_local(2) == 0 then
                chdir = chdir * -1
            end
            e.facing = e.facing + chdir

            x, y, poly = get_offset_location(e.x, e.y, e.z, e.polygon, e.facing, e._horizontal_dist)

            if poly then
                e._horizontal_dist = e._horizontal_dist / 6
                if e._vertical_velocity > 0 then
                    e._vertical_velocity = e._vertical_velocity / 2
                end
            end
        end

        if not poly then
            x, y, poly = e.x, e.y, e.polygon
        end

        if poly then
            e:position(x, y, z, poly)
        end

        return e
    elseif e._vertical_velocity <= -.09 then
        e._vertical_velocity = .03
        return e
    else
        e._active = false
        return nil
    end
end

function create_ephemera(x, y, z, polygon, col, seq, facing)
    if ephemera_count >= max_ephemera then
        cleanup()
    end

    local e = Ephemera.new(x, y, z, polygon, col, seq, facing)
    if not e then
        cleanup()
        e = Ephemera.new(x, y, z, polygon, col, seq, facing)
    end
    if e then
        ephemera_count = ephemera_count + 1
        e._owner = SP_NETGAME_OWNER
    end
    return e
end

function cleanup_ephemera_in_view()
    local rendered_ephemera = {}
    for e in Ephemera() do
        if e._owner == SP_NETGAME_OWNER and e.rendered and not e._active then
            table.insert(rendered_ephemera, e)
        end
    end

    -- Players.print("Rendered: " .. #rendered_ephemera .. ", total: " .. ephemera_count)
    if #rendered_ephemera > MAX_RENDERED_COUNT then
        cleanup(rendered_ephemera)
    end
end

function cleanup(ephemera)
    local count = 0

    if ephemera then
        for i = #ephemera, 1, -1 do
            local e = ephemera[i]
            if count % 3 == 0 and e and e._owner == SP_NETGAME_OWNER and not e._active then
                e:delete()
                ephemera_count = ephemera_count - 1
            end
            count = count + 1
        end
    else
        for e in Ephemera() do
            if count % 3 == 0 and e and e._owner == SP_NETGAME_OWNER and not e._active then
                e:delete()
                ephemera_count = ephemera_count - 1
            end
            count = count + 1
        end
    end
end

function get_ephemera_by_monster(m)
    local col = 6
    local seq = 35 + Game.random_local(3)
    local ct = get_color_table_from_monster(m.type)
    local num = 8

    return col, seq, ct, num
end

function get_color_table_from_monster(type)
    if type.mnemonic == "major tick" then
        return 1
    elseif type.mnemonic == "minor cyborg" then
        return 5
    elseif type.mnemonic == "minor tick" then
        return 5
    elseif type.mnemonic == "tiny yeti" then
        return nil
    elseif type.mnemonic == "big minor drone" then
        return 3
    elseif type.mnemonic == "big major drone" then
        return 7
    elseif type.mnemonic == "kamikaze tick" then
        return 4
    elseif type.mnemonic == "major cyborg" then
        return 0
    elseif type.mnemonic == "tiny pfhor" then
        return 2
    elseif type.mnemonic == "possessed drone" then
        return 6
    else
        return 0
    end
end

function get_ephemera_by_flaming_death()
    local col = 4
    local seq = 30 + Game.random_local(4)
    local ct = 0

    return col, seq, ct
end
